/***
 * Copyright (C) 1997,98 -- Miroslav Zagorac  <zaga@fly.cc.fer.hr>
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/param.h>

#ifdef __DJGPP__
#  include <pc.h>

#  define OUTB(a,b)   outb (b, a)
#else
#  include <linux/keyboard.h>

#if defined (__GLIBC__) && (__GLIBC__ >= 2)
#  include <sys/io.h>
#else
#  include <asm/io.h>
#endif /* __GLIBC__ */

#  define OUTB(a,b)   outb (a, b)
#endif /* __DJGPP__ */

#ifdef USE_MOUSE
#  include "lib_svgamouse.h"
#endif

#include "lib_debug.h"
#include "lib_pcx.h"
#include "lib_svga.h"
#include "lib_string.h"

#ifdef USE_JPEGLIB
#  include "lib_jpeg.h"
#endif


static palette_t        palette[SIZEOF_PALETTE];
static GraphicsContext *screen_gc = NULL;
static unchar          *vbuf = NULL;
static short            svga_flags = 0;

struct FONT_DATA svga_default_font;


#ifdef ANSI_FUNC

static inline short svga_ioperm (void)
#else

static inline short svga_ioperm ()
#endif
{
#ifdef __DJGPP__
   return (FUNC_RET_OK);
#else
   return (ioperm (0x3b4, 0x3df - 0x3b4 + 1, 1) ? FUNC_RET_ERROR : FUNC_RET_OK);
#endif
}


#ifdef DEBUG

#ifdef ANSI_FUNC

void svga_show_keytable (void)
#else

void svga_show_keytable ()
#endif
{
   char  *keytable;
   short  i;

   keytable = keyboard_getstate ();

   for (i = 0; i < NR_KEYS; i++)
      if (keytable[i])
      {
         /***
          * Nek' se zna da sam ovo protiv svoje volje stavio bas tu. B)
          */
         debug (("{%s()", __FUNCTION__))

         for (i = 0; i < NR_KEYS; i++)
            debug (("%s%d%c", ((i + 1) % 32) ? "" : "", (int) keytable[i], ((i + 1) % 32) ? ' ' : '\n'))

         debug_noreturn
      }
}


#ifdef ANSI_FUNC

void svga_show_palette (short x, short y, short block_size)
#else

void svga_show_palette (x, y, block_size)
   short x;
   short y;
   short block_size;
#endif
{
   short i, j;

   debug (("{%s(%d, %d, %d)\n", __FUNCTION__, x, y, block_size))

   for (i = 0; i < 16; i++)
      for (j = 0; j < 16; j++)
         gl_fillbox (x + j * 18, y + i * 18, block_size, block_size, i * 16 + j);

   debug_endfunc
}

#endif


#ifdef ANSI_FUNC

short svga_virtual_init (short mode, short buffering_flag)
#else

short svga_virtual_init (mode, buffering_flag)
   short mode;
   short buffering_flag;
#endif
/***
 * Nakon poziva ove funkcije, bez obzira da li je ukljuceno buferiranje, treba voditi
 * obzira o slijedecem:
 *
 * - poziv neke gl_funkcije koja nesto crta po ekranu, ne vodi do toga da se nacrtano i vidi.
 * - da bi se nacrtana slika vidjela na ekranu treba pozvati funkciju svga_copyscreen().
 * - pri velikim rezolucijama to moze dovesti do znatnog smanjivanja broja frameova.
 * - funkcija alocira polje u memoriji dovoljno veliko da u njega smjesti jedan ekran.
 *
 * primjer koristenja:
 *
 * main ()
 * {
 *    if (svga_init (VGAMODE, 0) != FUNC_RET_OK)
 *       perror_exit ("Graphics mode not supported", EX_SOFTWARE);
 * 
 *    if (svga_virtual_init (VGAMODE, 1) != FUNC_RET_OK)
 *       perror_exit ("Cannot initialize virtual buffers", EX_SOFTWARE);
 * 
 *    while (1)
 *    {
 *       ... nesto se nacrta sa gl_() funkcijama ...
 * 
 *       svga_copyscreen ();
 *    }
 * }
 *
 * *** Vazno: treba pazito da se koriste gl_() funkcije umjesto vga_(), gdje god se to moze
 *            (recimo umjesto vga_drawline() koristi gl_line() itd...).
 */
{
   debug (("{%s(%d, %d)", __FUNCTION__, mode, buffering_flag))

   if ((svga_flags & SVGA_OPEN) && !screen_gc)
   {
      short buffering;

      screen_gc = gl_allocatecontext ();
      gl_getcontext (screen_gc);
      gl_setcontextvga (mode);
      gl_setcontext (screen_gc);

      if (buffering_flag)
      {
         if ((buffering = gl_enablepageflipping (screen_gc)) < 2)
            debug_return ("} == %d", FUNC_RET_ERROR)

         svga_flags |= (buffering == 2) ? SVGA_DOUBLE_BUFFERING : SVGA_TRIPLE_BUFFERING;

         debug (("buffering = %d\n", buffering))
      }

      /***
       * Create virtual screen...
       */
      if ((vbuf = calloc (vga_getxdim (), vga_getydim ())) == NULL)
         debug_return ("} == %d", FUNC_RET_ERROR)

      gl_setcontextvirtual (vga_getxdim (), vga_getydim (), 1, 8, vbuf);
      gl_setclippingwindow (0, 0, vga_getxdim () - 1, vga_getydim () - 1);
   }

   debug_return ("} == %d", FUNC_RET_OK)
}


#ifdef ANSI_FUNC

short svga_init (short mode, short lock_flag)
#else

short svga_init (mode, lock_flag)
   short mode;
   short lock_flag;
#endif
{
   debug (("{%s(%d, %d)", __FUNCTION__, mode, lock_flag))

   svga_default_font.width     = 8;
   svga_default_font.height    = 8;
   svga_default_font.writemode = WRITEMODE_MASKED | FONT_COMPRESSED;
   svga_default_font.bitmap    = gl_font8x8;

   if (svga_ioperm () != FUNC_RET_OK)
      debug_return ("} == %d", FUNC_RET_ERROR)

   vga_disabledriverreport ();

   if (vga_init ())
      debug_return ("} == %d", FUNC_RET_ERROR)

   if (!vga_hasmode (mode) || vga_setmode (mode))
      debug_return ("} == %d", FUNC_RET_ERROR)

   (void) vga_clear ();

   if (gl_setcontextvga (mode))
      debug_return ("} == %d", FUNC_RET_ERROR)

   vga_setdisplaystart (0);
   gl_setclippingwindow (0, 0, vga_getxdim () - 1, vga_getydim () - 1);

   /***
    * XXX  Ovdje cu pokusati prebaciti graficku karticu u 8bpp mod (ovo bi trebalo samo
    *      ako se koristi nekakav graficki mod sa 256 boja).
    */
   if (vga_ext_set (VGA_EXT_AVAILABLE, VGA_AVAIL_FLAGS) & VGA_CLUT8)
   {
      (void) vga_ext_set (VGA_EXT_SET, VGA_CLUT8);   /* 8 bits per color */

      svga_flags |= SVGA_8BPP_PALETTE;
   }

   svga_get_palette (palette);
   svga_palette_fade (0);
   (void) keyboard_init ();

   svga_flags |= SVGA_OPEN;

#ifndef __DJGPP__
   if (lock_flag)
      vga_lockvc ();
#endif

#ifdef USE_MOUSE
   if (svga_mouse_init ())
      debug_return ("} == %d", FUNC_RET_WARNING)
#endif

   debug_return ("} == %d", FUNC_RET_OK)
}


#ifdef ANSI_FUNC

void svga_clean (void)
#else

void svga_clean ()
#endif
{
   debug (("{%s()", __FUNCTION__))

   if (vbuf)
      free (vbuf);

   vbuf = NULL;

   image_clean ();

   debug_endfunc
}


#ifdef ANSI_FUNC

void svga_close (void)
#else

void svga_close ()
#endif
{
   debug (("{%s()", __FUNCTION__))

   /***
    * Ako SVGA graficki mod nije bio inicijaliziran, neka od ovih funkcija ce
    * prouzrokovati gresku, koja ce pozvati exit(), a on ce pozvati opet svga_close(),
    * i tako do u vjecnost... Zato postoji flag svga_flags, tako da se to ne desava.
    */
   if (svga_flags & SVGA_OPEN)
   {
#ifdef USE_MOUSE
      svga_mouse_close ();
#endif

      /***
       * svgalib ima nezgodnu naviku pozivati prebaciti graficku karticu u text mod,
       * prije nego sto to atexit() uradi, pa je slijedeci red samo jos jedna gnjavaza...
       */
      svga_palette_fade (-2);
      svga_clean ();
      keyboard_close ();
      (void) vga_setmode (TEXT);

#ifndef __DJGPP__
      vga_unlockvc ();
#endif

      svga_flags &= ~SVGA_OPEN;
   }

   debug_endfunc
}


#ifdef ANSI_FUNC

palette_t *svga_palette (void)
#else

palette_t *svga_palette ()
#endif
{
   return (palette);
}


#ifdef ANSI_FUNC

GraphicsContext *svga_gc (void)
#else

GraphicsContext *svga_gc ()
#endif
{
   return (screen_gc);
}


#ifdef ANSI_FUNC

void svga_get_palette (palette_t *palette_data)
#else

void svga_get_palette (palette_data)
   palette_t *palette_data;
#endif
{
   short i;

   debug (("{%s(0x%08lx)", __FUNCTION__, (long) palette_data))

   /***
    * Ovdje sam mogao koristiti i gl_getpalette(), ali onda paleta ne bi uvijek bila 8bpp.
    */
   for (i = 0; i < SIZEOF_PALETTE; i += 3)
   {
      int r, g, b;

      (void) vga_getpalette (i / 3, &r, &g, &b);

      palette_data[i]     = r << BPP_SHIFT;
      palette_data[i + 1] = g << BPP_SHIFT;
      palette_data[i + 2] = b << BPP_SHIFT;
   }

   debug_endfunc
}


#ifdef ANSI_FUNC

void svga_set_palette (palette_t *palette_data, short copy_flag)
#else

void svga_set_palette (palette_data, copy_flag)
   palette_t *palette_data;
   short copy_flag;
#endif
{
   short i;

   debug (("{%s(0x%08lx)", __FUNCTION__, (long) palette_data))

   /***
    * Ovdje sam mogao koristiti i gl_setpalette(), ali onda paleta ne bi uvijek bila 8bpp.
    */
   for (i = 0; i < SIZEOF_PALETTE; i += 3)
      (void) vga_setpalette (i / 3, palette_data[i] >> BPP_SHIFT, palette_data[i + 1] >> BPP_SHIFT, palette_data[i + 2] >> BPP_SHIFT);

   if (copy_flag)
      (void) memcpy (palette, palette_data, SIZEOF_PALETTE);

   debug_endfunc
}


#ifdef ANSI_FUNC

void svga_palette_fade (short direction)
#else

void svga_palette_fade (direction)
   short direction;
#endif
{
   short i;

   debug (("{%s(%d)", __FUNCTION__, direction))

   if (direction)
   {
      palette_t tmp_palette[SIZEOF_PALETTE];
      short     start      = (direction > 0) ? 0 : BPP_MAXVALUE;
      short     stop       = (direction > 0) ? BPP_MAXVALUE : 0;

      for ( ; (direction > 0) ? (start < stop) : (start > stop); start += direction)
      {
         /***
          * (start / BPP_MAXVALUE) je vrijednost izmedju 0.0 i 1.0.
          */
         for (i = 0; i < SIZEOF_PALETTE; i++)
            tmp_palette[i] = start * palette[i] / BPP_MAXVALUE;

         vga_waitretrace ();
         svga_set_palette (tmp_palette, 0);
      }
   }
   else
   {
      /***
       * Sve boje u paleti se postave na 0 (crna).
       */
      for (i = 0; i < SIZEOF_PALETTE; i += 3)
         (void) vga_setpalette (i / 3, 0, 0, 0);
   }

   debug_endfunc
}


#ifdef ANSI_FUNC

void svga_set_border (unchar color)
#else

void svga_set_border (color)
   unchar color;
#endif
{
   int vgaIOBase;

   debug (("{%s(%d)", __FUNCTION__, color))

   (void) svga_ioperm ();

   vgaIOBase = (inb (0x03cc) & 0x01) ? 0x03d0 : 0x03b0;

   inb (ATR_CTL_INDEX_DATA_SWITCH);
   OUTB (0x11, ATR_CTL_INDEX);
   OUTB (color, ATR_CTL_DATA_W);
   inb (ATR_CTL_INDEX_DATA_SWITCH);
   OUTB (0x11 | 0x20, ATR_CTL_INDEX);
   OUTB (color, ATR_CTL_DATA_W);

   debug_endfunc
}


#ifdef ANSI_FUNC

void svga_screen_on (void)
#else

void svga_screen_on ()
#endif
{
   debug (("{%s()", __FUNCTION__))

#if 1
   (void) svga_ioperm ();

   /***
    * Turn screen back on
    */
   OUTB (0x01, SEQ_I);
   OUTB (inb (SEQ_D) & 0xDF, SEQ_D);

   /***
    * Enable video output
    */
   (void) inb (IS1_RC);
   OUTB (0x20, ATT_IW);
#else
   (void) vga_screenon ()
#endif

   debug_endfunc
}


#ifdef ANSI_FUNC

void svga_screen_off (void)
#else

void svga_screen_off ()
#endif
{
   debug (("{%s()", __FUNCTION__))

#if 1
   (void) svga_ioperm ();

   /***
    * Turn screen off
    */
   OUTB (0x01, SEQ_I);
   OUTB (inb (SEQ_D) | 0x20, SEQ_D);

   /***
    * Disable video output
    */
   (void) inb (IS1_RC);
   OUTB (0x00, ATT_IW);
#else
   (void) vga_screenoff ()
#endif

   debug_endfunc
}


#ifdef ANSI_FUNC

short svga_show_image (char *filename, short x, short y, short width, short height, short x1, short y1)
#else

short svga_show_image (filename, x, y, width, height, x1, y1)
   char *filename;
   short x;
   short y;
   short width;
   short height;
   short x1;
   short y1;
#endif
{
   struct IMAGE_DATA *image;

   debug (("{%s(\"%s\", %d, %d, %d, %d, %d, %d)", __FUNCTION__, filename, x, y, width, height, x1, y1))

   if ((image = image_load (filename, NULL, 0)))
   {
      /***
       * Ako width i height nisu zadani, uzimaju se iz slike...
       */
      gl_putboxpart (x, y, width ? width : image->width, height ? height : image->height, image->width, image->height, image->image, x1, y1);
      image_free (&image);

      debug_return ("} == %d", FUNC_RET_OK)
   }

   debug_return ("} == %d", FUNC_RET_ERROR)
}


#ifdef ANSI_FUNC

void svga_draw_rectangle (short x, short y, short width, short height, short color)
#else

void svga_draw_rectangle (x, y, width, height, color)
   short x;
   short y;
   short width;
   short height;
   short color;
#endif
{
   short x2 = x + width - 1, y2 = y + height - 1;

   debug (("{%s(%d, %d, %d, %d, %d)", __FUNCTION__, x, y, width, height, color))

   gl_hline (x, y, x2, color);
   gl_hline (x, y2, x2, color);
   gl_line (x, y, x, y2, color);
   gl_line (x2, y, x2, y2, color);

   debug_endfunc
}


#ifdef ANSI_FUNC

short svga_save_snapshot (void)
#else
   
short svga_save_snapshot ()
#endif
{
   struct IMAGE_DATA *snapshot;

   debug (("{%s()", __FUNCTION__))

   if ((snapshot = image_alloc (vga_getxdim (), vga_getydim ())) != NULL)
   {
      char filename[MAXPATHLEN];

      (void) sprintf (filename, "snapshot_%ld.pcx", (long) time (NULL));
      gl_getbox (0, 0, snapshot->width, snapshot->height, snapshot->image);
      (void) memcpy (snapshot->palette, palette, SIZEOF_PALETTE);
      (void) pcx_image_save (snapshot, filename);
      image_free (&snapshot);

      debug_return ("} == %d", FUNC_RET_OK)
   }

   debug_return ("} == %d", FUNC_RET_ERROR)
}


#ifdef ANSI_FUNC

void svga_copyscreen (short wait_retrace_flag)
#else
   
void svga_copyscreen (wait_retrace_flag)
   short wait_retrace_flag;
#endif
{
   /***
    * Nesto bi trebalo izmudrijati ovdje, jer SVGAlib ima glupi bug kod prebacivanja
    * stranica koje se prikazuju, tako da se povremeno vidi na ekranu brljotina umjesto
    * prave slike. I sad, u cemu je korist tripple bufferiranja ako ja tu cekam
    * retrace signal? U nicemu!
    */
   if (wait_retrace_flag)
      vga_waitretrace ();

   gl_copyscreen (screen_gc);
}


#ifdef ANSI_FUNC

void svga_printf (short x, short y, short color, struct FONT_DATA *font, char *format, ...)
#else

void svga_printf (x, y, color, font, format)
   short x;
   short y;
   short color;
   struct FONT_DATA *font;
   char *format; /* XXX  varargs */
#endif
/***
 * XXX  Prilikom ispisivanja teksta, pozadina moze biti samo prozirna ili crna.
 */
{
   va_list argptr;
   char    buffer[BUFSIZ];

   debug (("{%s(%d, %d, %d, 0x%08lx...)", __FUNCTION__, x, y, color, (long) font))

   if (! font)
      font = &svga_default_font;

   va_start (argptr, format);
   (void) vsprintf (buffer, format, argptr);
   va_end (argptr);

   debug (("font = {\n"))
   debug (("   width     = %d\n", font->width))
   debug (("   height    = %d\n", font->height))
   debug (("   writemode = %d\n", font->writemode))
   debug (("   bitmap    = 0x%08lx\n", (long) font->bitmap))
   debug ((" }\n"))

   gl_setwritemode (font->writemode);
   gl_setfont (font->width, font->height, font->bitmap);

   if (font->writemode & FONT_COMPRESSED)
   {
      gl_setfontcolors (BLACK, color);
      gl_write (x, y, buffer);
   }
   else
   {
      short i, font_charsize = font->width * font->height;

      for (i = 0; buffer[i]; i++)
      {
         font_char_color (font, color, buffer[i]);

         if (font->writemode & WRITEMODE_MASKED)
            gl_putboxmask (x + i * font->width, y, font->width, font->height, font->bitmap + buffer[i] * font_charsize);
         else
            gl_putbox (x + i * font->width, y, font->width, font->height, font->bitmap + buffer[i] * font_charsize);
      }
   }

   debug_endfunc
}
