/***
 * Copyright (C) 1996-98. by Miroslav Zagorac <zaga@tirnanog.cc.fer.hr>
 *
 * Funkcije su napisane na temelju odlicnih dokumenata koji se nalaze na
 * web stranicama: http://www.stack.nl/~ep/sp/peterb/index.html
 * Funkcija next_random() je preuzeta (i prepravljena) iz programa
 * decodeCaves.c, kojeg su autori:
 *   Jeff Bevis <bevis@ecn.purdue.edu>
 *   Peter Broadribb <peterb@perth.dialix.oz.au>
 *
 * Ovom prilikom se zahvaljujem gore spomenutim ljudima (sto oni naravno
 * nece shvatiti, jer ne znaju hrvatski B).
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <sys/types.h>

#include "../data/objects_def.h"
#include "../lib/lib_debug.h"

#include "cave.h"
#include "cave_data.h"
#include "graphics.h"
#include "level_editor.h"
#include "main.h"
#include "objects.h"
#include "commodore.h"
#include "spectrum.h"


static unchar *cave_area, *cave_attr;


#ifdef ANSI_FUNC

void store_single_object (unchar object, unchar column, unchar row)
#else

void store_single_object (object, column, row)
   unchar object;
   unchar column;
   unchar row;
#endif
{
   static struct
   {
      unchar object;
      unchar new_object;
      unchar move_direction;
   } conversion_table[] =
      {
         {   FIREFLY_1,      FIREFLY_1, 0 },
         {   FIREFLY_2,      FIREFLY_1, 1 },
         {   FIREFLY_3,      FIREFLY_1, 2 },
         {   FIREFLY_4,      FIREFLY_1, 3 },
         { BUTTERFLY_1,    BUTTERFLY_1, 2 },
         { BUTTERFLY_2,    BUTTERFLY_1, 3 },
         { BUTTERFLY_3,    BUTTERFLY_1, 0 },
         { BUTTERFLY_4,    BUTTERFLY_1, 1 },
         {  MAGIC_WALL, PRE_MAGIC_WALL, 0 },
         {        0xff,              0, 0 }
      };
   /***
    * Originalni nivoi pocinju na koordinati 0,2, a ja to pretvaram u 0,0.
    */
   short i, position = CAVE_POSITION (column, row - 2);

   debug (("{%s(%d, %d, %d)", __FUNCTION__, object, column, row))

   for (i = 0; conversion_table[i].object < 0xff; i++)
      if (object == conversion_table[i].object)
      {
         cave_attr[position] |= conversion_table[i].move_direction;
         object               = conversion_table[i].new_object;

         break;
      }

   cave_area[position] = object;

   debug_endfunc
}


#ifdef ANSI_FUNC

void draw_line (unchar object, unchar column, unchar row, unchar length, unchar direction)
#else

void draw_line (object, column, row, length, direction)
   unchar object;
   unchar column;
   unchar row;
   unchar length;
   unchar direction;
#endif
{
   static short direction_x[] = {  0,  1,  1,  1,  0, -1, -1, -1 };
   static short direction_y[] = { -1, -1,  0,  1,  1,  1,  0, -1 };
   short        i;

   debug (("{%s(%d, %d, %d, %d, %d)", __FUNCTION__, object, column, row, length, direction))

   for (i = 0; i < length; i++)
   {
      store_single_object (object, column, row);

      row    += direction_y[direction];
      column += direction_x[direction];
   }

   debug_endfunc
}


#ifdef ANSI_FUNC

void draw_filled_rectangle (unchar object, unchar column, unchar row, unchar length, unchar height, unchar object_fill)
#else

void draw_filled_rectangle (object, column, row, length, height, object_fill)
   unchar object;
   unchar column;
   unchar row;
   unchar length;
   unchar height;
   unchar object_fill;
#endif
{
   short i;

   debug (("{%s(%d, %d, %d, %d, %d, %d)", __FUNCTION__, object, column, row, length, height, object_fill))

   draw_rectangle (object, column, row, length, height);

   for (i = 1; i < (height - 1); i++)
      draw_line (object_fill, column + 1, row + i, length - 2, LINE_DIR_RIGHT);

   debug_endfunc
}


#ifdef ANSI_FUNC

void draw_rectangle (unchar object, unchar column, unchar row, unchar length, unchar height)
#else

void draw_rectangle (object, column, row, length, height)
   unchar object;
   unchar column;
   unchar row;
   unchar length;
   unchar height;
#endif
{
   debug (("{%s(%d, %d, %d, %d, %d)", __FUNCTION__, object, column, row, length, height))

   draw_line (object, column, row, length, LINE_DIR_RIGHT);
   draw_line (object, column, row + height - 1, length, LINE_DIR_RIGHT);
   draw_line (object, column, row + 1, height - 2, LINE_DIR_DOWN);
   draw_line (object, column + length - 1, row + 1, height - 2, LINE_DIR_DOWN);

   debug_endfunc
}


#ifdef ANSI_FUNC

void next_random (unchar *random_seed_1, unchar *random_seed_2)
#else

void next_random (random_seed_1, random_seed_2)
   unchar *random_seed_1;
   unchar *random_seed_2;
#endif
/***
 * This is the mathematical random number generator from the Commodore 64
 * implementation of Boulder Dash I. The 6510 disassembly is given in
 * comments, and the C translation follows.
 */
{
   unchar rand_1, rand_2;
   ushort result;

   debug (("{%s(%d, %d)", __FUNCTION__, *random_seed_1, *random_seed_2))

   /***
    * next_random:
    * 7085        LDA random_seed_1
    * 7087        ROR
    * 7088        ROR
    * 7089        AND #$80
    * 708B        STA rand_1
    *      
    * Note: ROR on the 6510 works like this:
    *   7-->6-->5-->4-->3-->2-->1-->0-->C
    *   ^                               |
    *   |_______________________________|
    * In other words, it's a nine-bit rotate. Thus it takes two RORs to shift
    * the low bit (bit zero) into the high bit (bit 7).
    */
   rand_1 = *random_seed_1 << 7;

   /***
    * 708E        LDA random_seed_2
    * 7090        ROR
    * 7091        AND #$7F
    * 7093        STA rand_2
    */
   rand_2 = *random_seed_2 >> 1;

   /***
    * 7096        LDA random_seed_2
    * 7098        ROR
    * 7099        ROR
    * 709A        AND #$80
    * 709C        CLC
    * 709D        ADC random_seed_2
    */
   result = ((*random_seed_2 << 7) & 0xff) + *random_seed_2;

   /***
    * 709F        ADC #$13
    * 70A1        STA random_seed_2
    */
   *random_seed_2 = result = (result & 0xff) + ((result & 0xff00) ? 1 : 0) + 0x13;

   /***
    * 70A3        LDA random_seed_1
    * 70A5        ADC rand_1
    */
   result = *random_seed_1 + ((result & 0xff00) ? 1 : 0) + rand_1;

   /***
    * 70A8        ADC rand_2
    * 70AB        STA random_seed_1
    * 70AD        RTS
    */
   *random_seed_1 = (result & 0xff) + ((result & 0xff00) ? 1 : 0) + rand_2;

   debug_endfunc
}


#ifdef ANSI_FUNC

short cave_decode (short room, short level, struct CAVE_DATA *cave_data)
#else

short cave_decode (room, level, cave_data)
   short room;
   short level;
   struct CAVE_DATA *cave_data;
#endif
{
   struct CAVE_RAW_DATA *cave_raw;
   ushort                pos = 0, i, x, y;
   unchar                random_seed_1 = 0, random_seed_2;

   debug (("{%s(%d, %d, 0x%08lx)", __FUNCTION__, room, level, (long) cave_data))

   if ((room < 0) || (room >= CAVE_MAXROOMEXT) || (level < 0) || (level >= CAVE_MAXLEVEL))
      debug_return ("} == %d", FUNC_RET_ERROR)

   if (TEST_OPTION (OPT_COMMODORE))
   {
      objects_data = commodore_objects_data;
      cave_raw     = (struct CAVE_RAW_DATA *) commodore_cave_raw_data[room + level / 5 * CAVE_MAXROOMEXT];

      if (! cave_raw->data[0])
         debug_return ("} == %d", FUNC_RET_ERROR)

      commodore_change_colors (cave_raw);
   }
   else if (TEST_OPTION (OPT_SPECTRUM))
   {
      objects_data = spectrum_objects_data;
      cave_raw     = (struct CAVE_RAW_DATA *) spectrum_cave_raw_data[room + level / 5 * CAVE_MAXROOMEXT];

      if (! cave_raw->data[0])
         debug_return ("} == %d", FUNC_RET_ERROR)

      spectrum_change_colors (cave_raw);
   }
   else
   {
      objects_data = default_objects_data;
      cave_raw     = (struct CAVE_RAW_DATA *) spectrum_cave_raw_data[room + level / 5 * CAVE_MAXROOMEXT];

      if (! cave_raw->data[0])
         debug_return ("} == %d", FUNC_RET_ERROR)
   }

   cave_data->room         = room;
   cave_data->level        = level;
   cave_data->wall_time    = cave_raw->wall_time * 1000000L;
   cave_data->amoeba_time  = (cave_raw->time[level % 5] - cave_raw->wall_time) * 1000000L;
   cave_data->dvalue_init  = cave_raw->dvalue_init;
   cave_data->dvalue_extra = cave_raw->dvalue_extra;
   cave_data->diamonds     = cave_raw->diamonds[level % 5];
   cave_data->time         = cave_raw->time[level % 5] * 1000000L;
   cave_data->color[0]     = cave_raw->bg1_color;
   cave_data->color[1]     = cave_raw->bg2_color;
   cave_data->color[2]     = cave_raw->fg_color;

   cave_attr = cave_data->attr;
   cave_area = cave_data->data;

   random_seed_2 = cave_raw->random[level % 5];

   (void) memset (cave_area, SPACE, CAVE_SIZE);

   /***
    * Decode the random cave data.
    */
   for (y = 3; y <= 23; y++)
      for (x = 0; x <= 39; x++)
      {
         unchar object = DIRT;

         next_random (&random_seed_1, &random_seed_2);

         for (i = 0; i <= 3; i++)
            if (random_seed_1 < cave_raw->object_probability[i])
               object = cave_raw->object_random[i];

         store_single_object (object, x, y);
      }

   /***
    * Decode the explicit cave data.
    */
   while (cave_raw->data[pos] != 0xff)
      switch ((cave_raw->data[pos] >> 6) & 0x03)
      {
         case 0x00 : store_single_object (cave_raw->data[pos] & 0x3f, cave_raw->data[pos + 1], cave_raw->data[pos + 2]);
                     pos += 3;

                     break;

         case 0x01 : draw_line (cave_raw->data[pos] & 0x3f, cave_raw->data[pos + 1], cave_raw->data[pos + 2], cave_raw->data[pos + 3], cave_raw->data[pos + 4]);
                     pos += 5;

                     break;

         case 0x02 : draw_filled_rectangle (cave_raw->data[pos] & 0x3f, cave_raw->data[pos + 1], cave_raw->data[pos + 2], cave_raw->data[pos + 3], cave_raw->data[pos + 4], cave_raw->data[pos + 5]);
                     pos += 6;

                     break;

         case 0x03 : draw_rectangle (cave_raw->data[pos] & 0x3f, cave_raw->data[pos + 1], cave_raw->data[pos + 2], cave_raw->data[pos + 3], cave_raw->data[pos + 4]);
                     pos += 5;
      }

   /***
    * Sad treba potraziti izlaz iz labirinta, jer bi ga kasnije mogao sakriti okvir.
    */
   for (i = 0; i < CAVE_SIZE; i++)
      if (cave_data->data[i] == PRE_OUTBOX)
      {
         cave_out.x = i % CAVE_WIDTH;
         cave_out.y = i / CAVE_WIDTH;

         break;
      }

   draw_rectangle (WALL, 0, 2, 40, 22);

   debug_return ("} == %d", FUNC_RET_OK)
}


#ifdef DEBUG

#ifdef ANSI_FUNC

static short cave_check (short room, short level, struct CAVE_RAW_DATA *cave_raw)
#else
   
static short cave_check (room, level, cave_raw)
   short room;
   short level;
   struct CAVE_RAW_DATA *cave_raw;
#endif
{
   ushort pos = 0, i, x, y;
   unchar random_seed_1 = 0, random_seed_2;

   debug (("{%s(%d, %d, 0x%08lx)", __FUNCTION__, room, level, (long) cave_raw))

   if (! cave_raw->data[0])
   {
      debug (("Cave %c:%s not defined!\n", code_room (room), code_level (level)))
      debug_return ("} == %d", FUNC_RET_ERROR)
   }

   debug (("room         = %c\n", code_room (room)))
   debug (("level        = %s\n", code_level (level)))
   debug (("wall_time    = %d\n", (uint) cave_raw->wall_time))
   debug (("amoeba_time  = %d\n", (uint) cave_raw->time[level % 5] - cave_raw->wall_time))
   debug (("dvalue_init  = %d\n", (uint) cave_raw->dvalue_init))
   debug (("dvalue_extra = %d\n", (uint) cave_raw->dvalue_extra))
   debug (("diamonds     = %d\n", (uint) cave_raw->diamonds[level % 5]))
   debug (("time         = %d\n", (uint) cave_raw->time[level % 5]))

   random_seed_2 = cave_raw->random[level % 5];

   /***
    * Decode the random cave data.
    */
   for (y = 3; y <= 23; y++)
      for (x = 0; x <= 39; x++)
      {
         unchar object = DIRT;

         debug_off ();
         next_random (&random_seed_1, &random_seed_2);
         debug_on ();

         for (i = 0; i <= 3; i++)
            if (random_seed_1 < cave_raw->object_probability[i])
               object = cave_raw->object_random[i];

         debug (("store_single_object (%d [%d, %d])\n", (uint) object, (uint) x, (uint) y))
      }

   /***
    * Decode the explicit cave data.
    */
   while (cave_raw->data[pos] != 0xff)
      switch ((cave_raw->data[pos] >> 6) & 0x03)
      {
         case 0x00 : debug (("store_single_object (%d [%d, %d])\n", (uint) cave_raw->data[pos] & 0x3f, (uint) cave_raw->data[pos + 1], (uint) cave_raw->data[pos + 2]))
                     pos += 3;

                     break;

         case 0x01 : debug (("draw_line (%d [%d, %d]-[%d, %d])\n", (uint) cave_raw->data[pos] & 0x3f, (uint) cave_raw->data[pos + 1], (uint) cave_raw->data[pos + 2], (uint) cave_raw->data[pos + 3], (uint) cave_raw->data[pos + 4]))
                     pos += 5;

                     break;

         case 0x02 : debug (("draw_filled_rectangle (%d [%d, %d; %d, %d] %d)\n", (uint) cave_raw->data[pos] & 0x3f, (uint) cave_raw->data[pos + 1], (uint) cave_raw->data[pos + 2], (uint) cave_raw->data[pos + 3], (uint) cave_raw->data[pos + 4], (uint) cave_raw->data[pos + 5]))
                     pos += 6;

                     break;

         case 0x03 : debug (("draw_rectangle (%d [%d, %d; %d, %d])\n", (uint) cave_raw->data[pos] & 0x3f, (uint) cave_raw->data[pos + 1], (uint) cave_raw->data[pos + 2], (uint) cave_raw->data[pos + 3], (uint) cave_raw->data[pos + 4]))
                     pos += 5;
      }

   debug (("Cave real size: %d bytes\n", pos + 0x21));

   for (i = pos; i < (CAVE_RAW_MAXSIZE - 0x21); i++)
      if (cave_raw->data[i])
         pos = i;

   debug (("Cave size: %d bytes\n", pos + 0x21))

   debug_return ("} == %d", FUNC_RET_OK)
}


#ifdef ANSI_FUNC

void caves_all_check (void)
#else

void caves_all_check ()
#endif
{
   short room, level;

   debug (("{%s()", __FUNCTION__))

   debug (("--- Commodore levels ---\n"))

   for (level = 0; level < 20/*CAVE_MAXLEVEL*/; level += 5)
      for (room = 0; room < CAVE_MAXROOMEXT; room++)
         (void) cave_check (room, level, (struct CAVE_RAW_DATA *) commodore_cave_raw_data[room + level / 5 * CAVE_MAXROOMEXT]);

   debug (("--- ZX Spectrum levels ---\n"))

   for (level = 0; level < 20/*CAVE_MAXLEVEL*/; level += 5)
      for (room = 0; room < CAVE_MAXROOMEXT; room++)
         (void) cave_check (room, level, (struct CAVE_RAW_DATA *) spectrum_cave_raw_data[room + level / 5 * CAVE_MAXROOMEXT]);

   debug_endfunc
}

#endif /* DEBUG */
