/***
 * Copyright (C) 1996-98. by Miroslav Zagorac <zaga@tirnanog.cc.fer.hr>
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>

#include "../lib/lib_debug.h"

#include "cave.h"
#include "cave_data.h"
#include "demo.h"
#include "graphics.h"
#include "keyboard.h"
#include "main.h"
#include "player.h"
#include "score.h"


struct COORDINATE rockford_coord;


#ifdef ANSI_FUNC

void player_data_set (short player, short room, short level)
#else

void player_data_set (player, room, level)
   short player;
   short room;
   short level;
#endif
{
   short i;

   debug (("{%s(%d, %d, %d)", __FUNCTION__, player, room, level))

   for (i = ((player < MAX_PLAYERS) ? player : 0); i < ((player < MAX_PLAYERS) ? (player + 1) : MAX_PLAYERS); i++)
   {
      player_data[i].room       = room;
      player_data[i].level      = level;
      player_data[i].lives      = 3;
      player_data[i].score      = 0L;
      player_data[i].last_bonus = 0L;
   }

   debug_endfunc
}


#ifdef ANSI_FUNC

short player_move (void)
#else

short player_move ()
#endif
{
   static short  old_x = 666, old_y = 666;
   static unchar rockford_run = ROCKFORD_R;
   short         rockford_pos = CAVE_POSITION (rockford_coord.x, rockford_coord.y), rockford_pos_new, is_out = 0, x, y;
   unchar        rockford = cave.data[rockford_pos];

   debug (("{%s()", __FUNCTION__))

   if (key_pressed_code (KEY_SPACE, 0))
   {
      game_data.game_status |= GAME_PAUSE;

      debug_return ("} == %d", 0)
   }
   else if (game_data.game_status & GAME_DEMO_MODE)
   {
      short demo_key = demo_keyboard (0);

      x = (demo_key == KEY_GAME_LEFT) ? -1 : ((demo_key == KEY_GAME_RIGHT) ? 1 : 0);
      y = x ? 0 : ((demo_key == KEY_GAME_UP) ? -1 : ((demo_key == KEY_GAME_DOWN) ? 1 : 0));
   }
   else
   {
      /***
       * Testiranje pritiska na neku od tipaka...
       */
      x = key_pressed_code (KEY_GAME_LEFT, 1) ? -1 : (key_pressed_code (KEY_GAME_RIGHT, 1) ? 1 : 0);
      y = x ? 0 : (key_pressed_code (KEY_GAME_UP, 1) ? -1 : (key_pressed_code (KEY_GAME_DOWN, 1) ? 1 : 0));
   }

#ifdef DEBUG
   if (key_pressed_code (KEY_SNAPSHOT, 0))
      (void) graph_save_snapshot ();
#endif

   rockford_pos_new = rockford_pos + CAVE_POSITION (x, y);

   /***
    * Promjena slike objekta (Rockford)
    */
   if (old_x != x)
   {
      if (!! (old_x = x))
         rockford_run = rockford = (x < 0) ? ROCKFORD_L : ROCKFORD_R;
   }
   else if (old_y != y)
   {
      if (!! (old_y = y))
         rockford = rockford_run;
   }

   if (!(game_data.game_status & GAME_DEMO_MODE) && (key_pressed_code (KEY_GAME_FIRE, 1)))
   {
      /***
       * Pritisnuta je tipka FIRE...
       */
      if (cave.data[rockford_pos_new] == DIAMOND)
      {
         cave_set_position (rockford_pos_new, SPACE, 0, 0);

         game_data.diamonds++;
         score_add (cave.dvalue_init);
      }
      else if (cave.data[rockford_pos_new] == BOULDER)
      {
         /***
          * Da se gurne kamen, treba i malo srece.
          */
         if (x && !y && (! (random () & 0x03)) && (cave.data[rockford_pos_new + x] == SPACE))
            cave_exchange_position (rockford_pos_new + x, rockford_pos_new);
      }
      else if (cave.data[rockford_pos_new] == DIRT)
         cave_set_position (rockford_pos_new, SPACE, 0, 0);

      /***
       * Ako nije pritisnuta neka od tipaka za smijer onda Rockford stoji, inace "hoda"
       * u smijeru u kojem je zadnji put isao.
       */
      if (x || y)
         cave.data[rockford_pos] = rockford_run;
      else if ((cave.data[rockford_pos] == ROCKFORD_R) || (cave.data[rockford_pos] == ROCKFORD_L))
         cave_set_position (rockford_pos, ROCKFORD, 0, 0);
   }
   else if (x || y)
   {
      static short is_pushed = 0;
      short        is_moved = 1;

      if ((cave.data[rockford_pos_new] == SPACE) || (cave.data[rockford_pos_new] == DIRT))
         is_pushed = 0;
      else if (cave.data[rockford_pos_new] == DIAMOND)
      {
         is_pushed = 0;

         game_data.diamonds++;
         score_add (cave.dvalue_init);
      }
      else if (cave.data[rockford_pos_new] == BOULDER)
      {
         /***
          * Prvo guranje kamena ide tesko (25%), a nakon toga malo lakse (75%).
          */
         if (x && !y && (is_pushed ? (random () & 0x03) : !(random () & 0x03)) && (cave.data[rockford_pos_new + x] == SPACE))
         {
            cave_exchange_position (rockford_pos_new + x, rockford_pos_new);

            is_pushed = 1;
         }
         else
            is_moved = 0;
      }
      else if (cave.data[rockford_pos_new] == FLASH_OUTBOX)
         is_out = 1;
      else
         is_moved = 0;

      if (is_moved)
      {
         cave_set_position (rockford_pos_new, rockford, 0, cave.anim[rockford_pos]);
         cave_set_position (rockford_pos, SPACE, 0, 0);

         rockford_coord.x += x;
         rockford_coord.y += y;
      }
      else
         cave.data[rockford_pos] = rockford;
   }
   else if ((cave.data[rockford_pos] == ROCKFORD_R) || (cave.data[rockford_pos] == ROCKFORD_L))
      cave_set_position (rockford_pos, ROCKFORD, 0, 0);

   debug_return ("} == %d", is_out)
}
